# chrono - agendas system
# Copyright (C) 2016-2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import itertools

from django.contrib.auth.models import Group
from django.db import transaction
from django.db.models import Q

from chrono.agendas.models import AbsenceReasonGroup, Agenda, AgendaImportError, UnavailabilityCalendar


def export_site(agendas=True, unavailability_calendars=True, absence_reason_groups=True):
    '''Dump site objects to JSON-dumpable dictionnary'''
    data = collections.OrderedDict()
    if absence_reason_groups:
        data['absence_reason_groups'] = [x.export_json() for x in AbsenceReasonGroup.objects.all()]
    if unavailability_calendars:
        data['unavailability_calendars'] = [x.export_json() for x in UnavailabilityCalendar.objects.all()]
    if agendas:
        qs1 = Agenda.objects.filter(~Q(kind='virtual'))
        qs2 = Agenda.objects.filter(kind='virtual')
        data['agendas'] = [x.export_json() for x in itertools.chain(qs1, qs2)]
    return data


def import_site(data, if_empty=False, clean=False, overwrite=False):
    if if_empty and (
        Agenda.objects.exists()
        or UnavailabilityCalendar.objects.exists()
        or AbsenceReasonGroup.objects.exists()
    ):
        return

    if clean:
        Agenda.objects.all().delete()
        UnavailabilityCalendar.objects.all().delete()
        AbsenceReasonGroup.objects.all().delete()

    results = {
        'agendas': {'created': 0, 'updated': 0},
        'unavailability_calendars': {'created': 0, 'updated': 0},
        'absence_reason_groups': {'created': 0, 'updated': 0},
    }
    agendas = data.get('agendas', [])
    unavailability_calendars = data.get('unavailability_calendars', [])
    absence_reason_groups = data.get('absence_reason_groups', [])

    role_names = set()
    for objs in (agendas, unavailability_calendars):
        role_names = role_names.union(
            {name for data in objs for _, name in data.get('permissions', {}).items() if name}
        )
    existing_roles = Group.objects.filter(name__in=role_names)

    if existing_roles.count() != len(role_names):
        existing_roles_names = set(existing_roles.values_list('name', flat=True))
        raise AgendaImportError('Missing roles: "%s"' % ', '.join(role_names - existing_roles_names))

    with transaction.atomic():
        for objs, cls, label in (
            (absence_reason_groups, AbsenceReasonGroup, 'absence_reason_groups'),
            (unavailability_calendars, UnavailabilityCalendar, 'unavailability_calendars'),
            (agendas, Agenda, 'agendas'),
        ):
            for data in objs:
                created = cls.import_json(data, overwrite=overwrite)
                if created:
                    results[label]['created'] += 1
                else:
                    results[label]['updated'] += 1
    return results
